"""Class definitions of the payloads used to send a media to ``hub``."""

from .data import TableReportTest, TableReportTrain
from .inspection import (
    Coefficients,
    ImpurityDecrease,
    PermutationImportanceTest,
    PermutationImportanceTrain,
)
from .model import EstimatorHtmlRepr
from .performance import (
    PrecisionRecallTest,
    PrecisionRecallTrain,
    PredictionErrorTest,
    PredictionErrorTrain,
    RocTest,
    RocTrain,
)

__all__ = [
    "Coefficients",
    "EstimatorHtmlRepr",
    "ImpurityDecrease",
    "PermutationImportanceTest",
    "PermutationImportanceTrain",
    "PrecisionRecallTest",
    "PrecisionRecallTrain",
    "PredictionErrorTest",
    "PredictionErrorTrain",
    "RocTest",
    "RocTrain",
    "TableReportTest",
    "TableReportTrain",
]



"""Class definition of the payload used to send a RMSE metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric


class Rmse(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.rmse"]] = "metrics.rmse"
    name: Literal["rmse"] = "rmse"
    verbose_name: Literal["RMSE"] = "RMSE"
    greater_is_better: Literal[False] = False
    position: Literal[3] = 3


class RmseTrain(Rmse):  # noqa: D101
    data_source: Literal["train"] = "train"


class RmseTest(Rmse):  # noqa: D101
    data_source: Literal["test"] = "test"


class RmseMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.rmse"]] = "metrics.rmse"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: Literal["rmse_mean"] = "rmse_mean"
    verbose_name: Literal["RMSE - MEAN"] = "RMSE - MEAN"
    greater_is_better: Literal[False] = False
    position: Literal[3] = 3


class RmseTrainMean(RmseMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class RmseTestMean(RmseMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class RmseStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.rmse"]] = "metrics.rmse"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: Literal["rmse_std"] = "rmse_std"
    verbose_name: Literal["RMSE - STD"] = "RMSE - STD"
    greater_is_better: Literal[False] = False
    position: None = None


class RmseTrainStd(RmseStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class RmseTestStd(RmseStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Definition of the payload used to associate inspection media with report."""

from __future__ import annotations

from abc import ABC
from collections.abc import Callable
from functools import reduce
from typing import ClassVar, Literal, cast

from orjson import OPT_NON_STR_KEYS, OPT_SERIALIZE_NUMPY, dumps

from skore_hub_project.artifact.media.media import Media, Report
from skore_hub_project.protocol import Display, EstimatorReport


class Inspection(Media[Report], ABC):  # noqa: D101
    accessor: ClassVar[str]
    content_type: Literal["application/vnd.dataframe"] = "application/vnd.dataframe"

    def content_to_upload(self) -> bytes | None:  # noqa: D102
        try:
            function = cast(
                "Callable[..., Display]",
                reduce(getattr, self.accessor.split("."), self.report),
            )
        except AttributeError:
            return None

        display = function()
        frame = display.frame()

        return dumps(
            frame.fillna("NaN").to_dict(orient="tight"),
            option=(OPT_NON_STR_KEYS | OPT_SERIALIZE_NUMPY),
        )


class PermutationImportance(Inspection[EstimatorReport], ABC):  # noqa: D101
    accessor: ClassVar[str] = "inspection.permutation_importance"
    name: Literal["permutation_importance"] = "permutation_importance"

    def content_to_upload(self) -> bytes | None:  # noqa: D102
        for key, display in reversed(list(self.report._cache.items())):
            if len(key) < 6:
                continue

            parent_hash, name, data_source, at_step, _, metric, *_ = key

            if (
                parent_hash == self.report._hash
                and name == "permutation_importance"
                and data_source == self.data_source
                and at_step == 0
                and metric is None
            ):
                frame = display.frame()

                return dumps(
                    frame.fillna("NaN").to_dict(orient="tight"),
                    option=(OPT_NON_STR_KEYS | OPT_SERIALIZE_NUMPY),
                )

        return None


class PermutationImportanceTrain(PermutationImportance):  # noqa: D101
    data_source: Literal["train"] = "train"


class PermutationImportanceTest(PermutationImportance):  # noqa: D101
    data_source: Literal["test"] = "test"


class ImpurityDecrease(Inspection[EstimatorReport]):  # noqa: D101
    accessor: ClassVar[str] = "inspection.impurity_decrease"
    name: Literal["impurity_decrease"] = "impurity_decrease"
    data_source: None = None


class Coefficients(Inspection[Report]):  # noqa: D101
    accessor: ClassVar[str] = "inspection.coefficients"
    name: Literal["coefficients"] = "coefficients"
    data_source: None = None



"""Class definition of the payload used to associate a media with the report."""

from abc import ABC
from typing import Generic, TypeVar

from pydantic import Field

from skore_hub_project.artifact.artifact import Artifact
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

Report = TypeVar("Report", bound=(EstimatorReport | CrossValidationReport))


class Media(Artifact, ABC, Generic[Report]):
    """
    Payload used to associate a media with the report.

    Attributes
    ----------
    project : Project
        The project to which the artifact's payload must be associated.
    content_type : str
        The content-type of the artifact content.
    report : EstimatorReport | CrossValidationReport
        The report on which compute the media.
    name : str
        The name of the media.
    data_source : str | None
        The source of the data used to generate the media.
    """

    report: Report = Field(repr=False, exclude=True)
    name: str = Field(init=False)
    data_source: str | None = Field(init=False)



"""Class definition of the payload used to send a recall metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric, cast_to_float


class Recall(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.recall"
    name: str = "recall"
    verbose_name: str = "Recall (macro)"
    greater_is_better: bool = True
    position: None = None

    def compute(self) -> None:
        """Compute the value of the metric."""
        try:
            function = self.report.metrics.recall
        except AttributeError:
            self.value = None
        else:
            self.value = cast_to_float(
                function(data_source=self.data_source, average="macro")
            )


class RecallTrain(Recall):  # noqa: D101
    data_source: Literal["train"] = "train"


class RecallTest(Recall):  # noqa: D101
    data_source: Literal["test"] = "test"


class RecallMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.recall"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: str = "recall_mean"
    verbose_name: str = "Recall (macro) - MEAN"
    greater_is_better: bool = True
    position: None = None

    def compute(self) -> None:
        """Compute the value of the metric."""
        try:
            function = self.report.metrics.recall
        except AttributeError:
            self.value = None
        else:
            dataframe = function(
                data_source=self.data_source, aggregate=self.aggregate, average="macro"
            )
            self.value = cast_to_float(dataframe.iloc[0, 0])


class RecallTrainMean(RecallMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class RecallTestMean(RecallMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class RecallStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.recall"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: str = "recall_std"
    verbose_name: str = "Recall (macro) - STD"
    greater_is_better: bool = False
    position: None = None

    def compute(self) -> None:
        """Compute the value of the metric."""
        try:
            function = self.report.metrics.recall
        except AttributeError:
            self.value = None
        else:
            dataframe = function(
                data_source=self.data_source, aggregate=self.aggregate, average="macro"
            )
            self.value = cast_to_float(dataframe.iloc[0, 0])


class RecallTrainStd(RecallStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class RecallTestStd(RecallStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Login to ``skore hub``."""

from collections.abc import Callable
from logging import getLogger

from rich.align import Align
from rich.live import Live
from rich.panel import Panel

from skore_hub_project import console
from skore_hub_project.authentication.apikey import APIKey
from skore_hub_project.authentication.token import Token
from skore_hub_project.authentication.uri import URI

logger = getLogger(__name__)

#
# Global variable storing credentials used for authentication by the ``HUBClient``, both
# with an API key or a temporary token.
#
# By default, it is empty and must be initialized by the user by calling explicitly the
# function ``login``.
#
credentials: Callable[[], dict[str, str]] | None = None


def login(*, timeout: int = 600) -> None:
    """Login to ``skore hub``."""
    global credentials

    if credentials is not None:
        logger.debug(f"Already logged in {URI} with {type(credentials)}.")
        console.print(
            Panel(
                Align.center("Already logged in."),
                title="[cyan]Login to [b]Skore Hub",
                border_style="cyan",
                padding=1,
            )
        )

        return

    try:
        credentials = APIKey()
    except KeyError:
        with Live(console=console, auto_refresh=False) as live:
            credentials = Token(timeout=timeout, live=live)

            live.update(
                Panel(
                    Align.center(
                        "Successfully logged in, using [b]interactive authentication."
                    ),
                    title="[cyan]Login to [bold]Skore Hub",
                    border_style="cyan",
                    padding=1,
                )
            )
            live.refresh()
    else:
        console.print(
            Panel(
                Align.center("Successfully logged in, using [b]API key."),
                title="[cyan]Login to [bold]Skore Hub",
                border_style="cyan",
                padding=1,
            )
        )



"""Class definition of the ``skore`` hub project."""

from __future__ import annotations

import itertools
import re
import warnings
from collections.abc import Callable
from functools import wraps
from operator import itemgetter
from re import sub as substitute
from tempfile import TemporaryFile
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    ParamSpec,
    Protocol,
    TypedDict,
    TypeVar,
    runtime_checkable,
)
from unicodedata import normalize

import joblib
import orjson
from httpx import HTTPStatusError

from skore_hub_project.client.client import Client, HUBClient
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

P = ParamSpec("P")
R = TypeVar("R")


if TYPE_CHECKING:

    class Metadata(TypedDict):  # noqa: D101
        id: str
        key: str
        date: str
        learner: str
        ml_task: str
        report_type: str
        dataset: str
        rmse: float | None
        log_loss: float | None
        roc_auc: float | None
        fit_time: float | None
        predict_time: float | None
        rmse_mean: float | None
        log_loss_mean: float | None
        roc_auc_mean: float | None
        fit_time_mean: float | None
        predict_time_mean: float | None


def slugify(string: str) -> str:
    """
    Slugify string.

    The string must be lower-case and contain only ASCII letters, digits, and characters
    ``.``, ``-``, and ``_``.

    In order:
    - convert to ASCII and ignore characters in error,
    - replace characters that aren't alphanumerics, dots, dashes or underscores by dash,
    - convert repeated dots to single dots,
    - convert repeated dashes to single dash,
    - convert repeated underscores to single underscore,
    - strip leading and trailing dots, dashes, and underscores.
    """
    string = normalize("NFKD", string).encode("ascii", "ignore").decode("ascii")

    string = string.lower()
    string = substitute(r"[^\w.-]", "-", string)
    string = substitute(r"[.]+", ".", string)
    string = substitute(r"[-]+", "-", string)
    string = substitute(r"[_]+", "_", string)

    return string.strip(".-_")


def slugify_and_warn(string: str, type: Literal["workspace", "name"]) -> str:
    """Slugify workspace or name string, and warn if the result differs."""
    slug = slugify(string)

    if slug != string:
        warnings.warn(
            (
                (
                    f"Your project will be addressed under the '{slug}' workspace. "
                    "The workspace name must be lower-case and contain only ASCII "
                    "letters, digits, and characters '.', '-', and '_'."
                )
                if type == "workspace"
                else (
                    f"Your project will be created as '{slug}'. "
                    "The project name must be lower-case and contain only ASCII letters"
                    ", digits, and characters '.', '-', and '_'."
                )
            ),
            stacklevel=2,
        )

    return slug


def ensure_project_is_created(method: Callable[P, R]) -> Callable[P, R]:
    """Ensure project is created before executing any other operation."""

    @runtime_checkable
    class Project(Protocol):
        created: bool
        workspace: str
        name: str

    @wraps(method)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        project = args[0]

        assert isinstance(project, Project), "You can only wrap `Project` methods"

        if not project.created:
            with HUBClient() as hub_client:
                hub_client.post(f"projects/{project.workspace}/{project.name}")

            project.created = True

        return method(*args, **kwargs)

    return wrapper


class Project:
    """
    API to manage a collection of key-report pairs persisted in a hub storage.

    It communicates with the Probabl's ``skore hub`` storage, based on the pickle
    representation. Its constructor initializes a hub project by creating a new
    project or by loading an existing one from a defined workspace.

    The class main methods are :func:`~skore_hub_project.Project.put`,
    :func:`~skore_hub_project.reports.metadata` and
    :func:`~skore_hub_project.reports.get`, respectively to insert a key-report pair
    into the Project, to obtain the reports metadata and to get a specific report.

    Parameters
    ----------
    workspace : str
        The workspace of the project.

        A workspace is a ``skore hub`` concept that must be configured on the
        ``skore hub`` interface. It represents an isolated entity managing users,
        projects, and resources. It can be a company, organization, or team that
        operates independently within the system.
    name : str
        The name of the project.

    Attributes
    ----------
    workspace : str
        The workspace of the project.
    name : str
        The name of the project.
    """

    __REPORT_URN_PATTERN = re.compile(
        r"skore:report:(?P<type>(estimator|cross-validation)):(?P<id>.+)"
    )

    def __init__(self, workspace: str, name: str):
        """
        Initialize a hub project.

        Initialize a hub project by creating a new project or by loading an existing
        one from a defined workspace.

        Parameters
        ----------
        workspace : Path
            The workspace of the project.

            A workspace is a ``skore hub`` concept that must be configured on the
            ``skore hub`` interface. It represents an isolated entity managing users,
            projects, and resources. It can be a company, organization, or team that
            operates independently within the system.
        name : str
            The name of the project.
        """
        self.created = False
        self.__workspace = slugify_and_warn(workspace, "workspace")
        self.__name = slugify_and_warn(name, "name")

    @property
    def workspace(self) -> str:
        """The workspace of the project."""
        return self.__workspace

    @property
    def name(self) -> str:
        """The name of the project."""
        return self.__name

    @ensure_project_is_created
    def put(self, key: str, report: EstimatorReport | CrossValidationReport) -> None:
        """
        Put a key-report pair to the hub project.

        If the key already exists, its last report is modified to point to this new
        report, while keeping track of the report history.

        Parameters
        ----------
        key : str
            The key to associate with ``report`` in the hub project.
        report : skore.EstimatorReport | skore.CrossValidationReport
            The report to associate with ``key`` in the hub project.

        Raises
        ------
        TypeError
            If the combination of parameters are not valid.
        """
        from ..report import CrossValidationReportPayload, EstimatorReportPayload

        if not isinstance(key, str):
            raise TypeError(f"Key must be a string (found '{type(key)}')")

        payload: EstimatorReportPayload | CrossValidationReportPayload

        if isinstance(report, EstimatorReport):
            payload = EstimatorReportPayload(project=self, key=key, report=report)
            endpoint = "estimator-reports"
        elif isinstance(report, CrossValidationReport):
            payload = CrossValidationReportPayload(project=self, key=key, report=report)
            endpoint = "cross-validation-reports"
        else:
            raise TypeError(
                f"Report must be a `skore.EstimatorReport` or "
                f"`skore.CrossValidationReport` (found '{type(report)}')"
            )

        payload_dict = payload.model_dump()
        payload_json_bytes = orjson.dumps(payload_dict, option=orjson.OPT_NON_STR_KEYS)

        with HUBClient() as hub_client:
            hub_client.post(
                url=f"projects/{self.workspace}/{self.name}/{endpoint}",
                content=payload_json_bytes,
                headers={
                    "Content-Length": str(len(payload_json_bytes)),
                    "Content-Type": "application/json",
                },
            )

    @ensure_project_is_created
    def get(self, urn: str) -> EstimatorReport | CrossValidationReport:
        """Get a persisted report by its URN."""
        if m := re.match(Project.__REPORT_URN_PATTERN, urn):
            workspace = self.workspace
            name = self.name
            type = m["type"]
            id = m["id"]
            url = f"projects/{workspace}/{name}/{type}-reports/{id}"
        else:
            raise ValueError(
                f"URN '{urn}' format does not match '{Project.__REPORT_URN_PATTERN}'"
            )

        # Retrieve presigned URL
        with HUBClient() as hub_client:
            response = hub_client.get(url=url)
            metadata = response.json()
            presigned_url = metadata["pickle"]["presigned_url"]

        report: EstimatorReport | CrossValidationReport

        # Download pickled report before unpickling it.
        #
        # It uses streaming responses that do not load the entire response body into
        # memory at once.
        with (
            TemporaryFile(mode="w+b") as tmpfile,
            Client() as client,
            client.stream(method="GET", url=presigned_url, timeout=30) as response,
        ):
            for data in response.iter_bytes():
                tmpfile.write(data)

            tmpfile.seek(0)

            report = joblib.load(tmpfile)

        return report

    @ensure_project_is_created
    def summarize(self) -> list[Metadata]:
        """Obtain metadata/metrics for all persisted reports in insertion order."""

        def dto(response: Any) -> Metadata:
            report_type, summary = response
            metrics = {
                metric["name"]: metric["value"]
                for metric in summary["metrics"]
                if metric["data_source"] in (None, "test")
            }

            return {
                "id": summary["urn"],
                "key": summary["key"],
                "date": summary["created_at"],
                "learner": summary["estimator_class_name"],
                "ml_task": summary["ml_task"],
                "report_type": report_type,
                "dataset": summary["dataset_fingerprint"],
                "rmse": metrics.get("rmse"),
                "log_loss": metrics.get("log_loss"),
                "roc_auc": metrics.get("roc_auc"),
                "fit_time": metrics.get("fit_time"),
                "predict_time": metrics.get("predict_time"),
                "rmse_mean": metrics.get("rmse_mean"),
                "log_loss_mean": metrics.get("log_loss_mean"),
                "roc_auc_mean": metrics.get("roc_auc_mean"),
                "fit_time_mean": metrics.get("fit_time_mean"),
                "predict_time_mean": metrics.get("predict_time_mean"),
            }

        with HUBClient() as hub_client:
            responses = itertools.chain(
                zip(
                    itertools.repeat("estimator"),
                    hub_client.get(
                        f"projects/{self.workspace}/{self.name}/estimator-reports/"
                    ).json(),
                ),
                zip(
                    itertools.repeat("cross-validation"),
                    hub_client.get(
                        f"projects/{self.workspace}/{self.name}/cross-validation-reports/"
                    ).json(),
                ),
            )

        return sorted(map(dto, responses), key=itemgetter("date"))

    def __repr__(self) -> str:  # noqa: D105
        return f"Project(mode='hub', name='{self.name}', workspace='{self.workspace}')"

    @staticmethod
    def delete(workspace: str, name: str) -> None:
        """
        Delete a hub project.

        Parameters
        ----------
        workspace : Path
            The workspace of the project.

            A workspace is a ``skore hub`` concept that must be configured on the
            ``skore hub`` interface. It represents an isolated entity managing users,
            projects, and resources. It can be a company, organization, or team that
            operates independently within the system.
        name : str
            The name of the project.
        """
        workspace = slugify_and_warn(workspace, "workspace")
        name = slugify_and_warn(name, "name")

        with HUBClient() as hub_client:
            try:
                hub_client.delete(f"projects/{workspace}/{name}")
            except HTTPStatusError as e:
                if e.response.status_code == 403:
                    raise PermissionError(
                        f"Failed to delete the project '{name}'; "
                        f"please contact the '{workspace}' owner"
                    ) from e
                raise



"""Class definition of the payload used to send a report to ``hub``."""

from __future__ import annotations

from abc import ABC
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import cached_property, partial
from typing import ClassVar, Generic, TypeVar, cast

from pydantic import BaseModel, ConfigDict, Field, computed_field
from rich.progress import BarColumn, Progress, TextColumn, TimeElapsedColumn

from skore_hub_project import switch_mpl_backend
from skore_hub_project.artifact.media.media import Media
from skore_hub_project.artifact.pickle import Pickle
from skore_hub_project.metric.metric import Metric
from skore_hub_project.project.project import Project
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

SkinnedProgress = partial(
    Progress,
    TextColumn("[bold cyan]{task.description}..."),
    BarColumn(
        complete_style="dark_orange",
        finished_style="dark_orange",
        pulse_style="orange1",
    ),
    TextColumn("[orange1]{task.percentage:>3.0f}%"),
    TimeElapsedColumn(),
    transient=True,
)

Report = TypeVar("Report", bound=(EstimatorReport | CrossValidationReport))


class ReportPayload(BaseModel, ABC, Generic[Report]):
    """
    Payload used to send a report to ``hub``.

    Attributes
    ----------
    METRICS : ClassVar[tuple[Metric, ...]]
        The metric classes that have to be computed from the report.
    MEDIAS : ClassVar[tuple[Media, ...]]
        The media classes that have to be computed from the report.
    project : Project
        The project to which the report payload should be sent.
    report : EstimatorReport | CrossValidationReport
        The report on which to calculate the payload to be sent.
    key : str
        The key to associate to the report.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    METRICS: ClassVar[tuple[type[Metric[Report]], ...]]
    MEDIAS: ClassVar[tuple[type[Media[Report]], ...]]

    project: Project = Field(repr=False, exclude=True)
    report: Report = Field(repr=False, exclude=True)
    key: str

    @computed_field  # type: ignore[prop-decorator]
    @property
    def estimator_class_name(self) -> str:
        """The name of the report's estimator."""
        return self.report.estimator_name_

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def dataset_fingerprint(self) -> str:
        """The hash of the targets in the test-set."""
        import joblib

        return cast(
            str,
            joblib.hash(
                self.report.y_test
                if isinstance(self.report, EstimatorReport)
                else self.report.y
            ),
        )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def ml_task(self) -> str:
        """The type of ML task covered by the report."""
        return self.report.ml_task

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def metrics(self) -> list[Metric[Report]]:
        """
        The list of scalar metrics that have been computed from the report.

        Notes
        -----
        Unavailable metrics have been filtered out.

        All metrics whose value is not a scalar are currently ignored:
        - ignore ``list[float]`` for multi-output ML task,
        - ignore ``dict[str: float]`` for multi-classes ML task.

        The position field is used to drive the ``hub``'s parallel coordinates plot:
        - int [0, inf[, to be displayed at the position,
        - None, not to be displayed.
        """
        self.report.cache_predictions()

        metrics = [metric_cls(report=self.report) for metric_cls in self.METRICS]

        with (
            switch_mpl_backend(),
            SkinnedProgress() as progress,
            ThreadPoolExecutor() as pool,
        ):
            tasks = [
                pool.submit(lambda metric: metric.compute(), metric)
                for metric in metrics
            ]

            for task in progress.track(
                as_completed(tasks),
                description=f"Computing {self.report.__class__.__name__} metrics",
                total=len(tasks),
            ):
                task.result()

        return [metric for metric in metrics if metric.value is not None]

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def medias(self) -> list[Media[Report]]:
        """
        The list of medias that have been computed from the report.

        Medias are `pandas.Dataframe`, SVG images, Python dictionaries or HTML string.

        Notes
        -----
        Unavailable medias have been filtered out.
        """
        payloads = []

        for media_cls in self.MEDIAS:
            payload = media_cls(project=self.project, report=self.report)

            if payload.checksum is not None:
                payloads.append(payload)

        return payloads

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def pickle(self) -> Pickle:
        """
        The checksum of the instance.

        The checksum of the instance that was assigned before being uploaded to the
        artifact storage. It is based on its ``joblib`` serialization and mainly used to
        retrieve it from the artifacts storage.
        """
        return Pickle(project=self.project, report=self.report)



"""Client exchanging with ``skore hub``."""

import importlib.metadata
from contextlib import suppress
from http import HTTPStatus
from logging import getLogger
from time import sleep
from typing import Any, Final
from urllib.parse import urljoin

from httpx import (
    URL,
    Headers,
    HTTPError,
    HTTPStatusError,
    NetworkError,
    RemoteProtocolError,
    Response,
    TimeoutException,
)
from httpx import Client as HTTPXClient
from httpx._types import HeaderTypes

from skore_hub_project.authentication.uri import URI

logger = getLogger(__name__)


class Client(HTTPXClient):
    """
    Client with a retry strategy based on exponential backoff algorithm.

    Parameters
    ----------
    retry : bool, optional
        Retry request on fails, default True.
    retry_total: int | None, optional
        Total number of retries to allow, default 10.
        Set to None to remove this constraint.
    retry_backoff_factor : float, optional
        A backoff factor to apply between retries, default 0.25.
    retry_backoff_max : float, optional
        Maximum backoff time, default 120.
    """

    RETRYABLE_STATUS_CODES: Final[frozenset[HTTPStatus]] = frozenset(
        (
            HTTPStatus.REQUEST_TIMEOUT,
            HTTPStatus.TOO_EARLY,
            HTTPStatus.TOO_MANY_REQUESTS,
            HTTPStatus.BAD_GATEWAY,
            HTTPStatus.SERVICE_UNAVAILABLE,
            HTTPStatus.GATEWAY_TIMEOUT,
        )
    )

    RETRYABLE_EXCEPTIONS: Final[tuple[type[HTTPError], ...]] = (
        TimeoutException,
        NetworkError,
        RemoteProtocolError,
    )

    STATUS_CLASS_TO_ERROR_TYPE: Final[dict[int, str]] = {
        1: "Informational response",
        3: "Redirect response",
        4: "Client error",
        5: "Server error",
    }

    def __init__(
        self,
        *,
        retry: bool = True,
        retry_total: int | None = 10,
        retry_backoff_factor: float = 0.25,
        retry_backoff_max: float = 120,
    ):
        super().__init__(follow_redirects=True, timeout=30)

        self.retry = retry
        self.retry_total = retry_total if retry_total is not None else float("inf")
        self.retry_backoff_factor = retry_backoff_factor
        self.retry_backoff_max = retry_backoff_max

    def request(self, *args: Any, **kwargs: Any) -> Response:
        """Execute request with retry strategy."""
        retries = 0

        while True:
            try:
                response = super().request(*args, **kwargs)
            except self.RETRYABLE_EXCEPTIONS:
                if not self.retry:
                    raise
            else:
                if response.is_success:
                    return response

                overlimit = retries >= self.retry_total
                unretryable = response.status_code not in self.RETRYABLE_STATUS_CODES

                if (not self.retry) or overlimit or unretryable:
                    break

            timeout = self.retry_backoff_factor * (2**retries)
            retries += 1

            sleep_duration = min(timeout, self.retry_backoff_max)

            logger.warning(
                "Request failed to reach server. "
                f"Trying again in {sleep_duration} seconds."
            )

            sleep(sleep_duration)

        # Raise extended exception with body message when available.
        status_class = response.status_code // 100
        error_type = self.STATUS_CLASS_TO_ERROR_TYPE.get(
            status_class, "Invalid status code"
        )

        message = (
            f"{error_type} '{response.status_code} {response.reason_phrase}' "
            f"for url '{response.url}'"
        )

        with suppress(Exception):
            message += f": {response.json()['message']}"

        raise HTTPStatusError(message, request=response.request, response=response)


def __semver(version: str) -> str | None:
    """
    Convert a PyPI version of the ``skore-hub-project`` package to semver.

    Parameters
    ----------
    version : str
        The PyPI version of the ``skore-hub-project`` package, that can only be:
        - ``0.0.0+unknown``
        - ``X.Y.Z``
        - ``X.Y.ZrcN``
    """
    if version == "0.0.0+unknown":
        return None

    return version.replace("rc", "-rc.")


PACKAGE_SEMVER = __semver(importlib.metadata.version("skore-hub-project"))


class HUBClient(Client):
    """Client exchanging with ``skore hub``."""

    def request(
        self,
        method: str,
        url: URL | str,
        headers: HeaderTypes | None = None,
        **kwargs: Any,
    ) -> Response:
        """Execute request with authorization."""
        from skore_hub_project.authentication.login import credentials

        if credentials is None:
            raise RuntimeError(
                "You are not logged in. "
                "Please call the `skore.login()` function at the top of your script."
            )

        headers = Headers(headers)

        # Overload headers with credentials
        headers.update(credentials())

        # Overload headers with package semantic versioning
        if PACKAGE_SEMVER:
            headers.update({"X-Skore-Client": f"skore-hub-project/{PACKAGE_SEMVER}"})

        # Prefix the request by the hub URI when ``url`` is not absolute
        url = urljoin(URI, str(url))

        return super().request(method=method, url=url, headers=headers, **kwargs)



"""Package that provides APIs to communicate between ``skore`` and ``skore hub``."""

import logging
from base64 import b64decode, b64encode
from collections.abc import Iterator
from contextlib import contextmanager

from rich.console import Console
from rich.theme import Theme

__all__ = [
    "Payload",
    "b64_str_to_bytes",
    "bytes_to_b64_str",
    "console",
    "switch_mpl_backend",
]


logging.basicConfig()
logger = logging.getLogger(__name__)

console = Console(
    width=88,
    theme=Theme(
        {
            "repr.str": "cyan",
            "rule.line": "orange1",
            "repr.url": "orange1",
        }
    ),
)


def b64_str_to_bytes(literal: str) -> bytes:
    """Decode the Base64 str object ``literal`` in a bytes."""
    return b64decode(literal.encode("utf-8"))


def bytes_to_b64_str(literal: bytes) -> str:
    """Encode the bytes-like object ``literal`` in a Base64 str."""
    return b64encode(literal).decode("utf-8")


@contextmanager
def switch_mpl_backend() -> Iterator[None]:
    """
    Context-manager for switching ``matplotlib.backend`` to ``agg``.

    Notes
    -----
    The ``agg`` backend is a non-interactive backend that can only write to files.
    It is used in ``skore-hub-project`` to generate artifacts where we don't need an
    X display.

    https://matplotlib.org/stable/users/explain/figure/backends.html#selecting-a-backend
    """
    import matplotlib

    original_backend = matplotlib.get_backend()

    try:
        matplotlib.use("agg")
        yield
    finally:
        matplotlib.use(original_backend)



"""API key used for ``skore hub`` authentication."""

from collections.abc import Callable
from os import environ
from typing import Final

ENV_VAR_NAME: Final[str] = "SKORE_HUB_API_KEY"


def APIKey() -> Callable[[], dict[str, str]]:
    """Get the API key used for ``skore hub`` authentication.

    In the form of HTTP header parameters.
    """
    header = {"X-API-Key": environ[ENV_VAR_NAME]}
    return lambda: header



"""URI used for ``skore hub`` authentication."""

from os import environ
from typing import Final

DEFAULT: Final[str] = "https://api.skore.probabl.ai"
ENV_VAR_NAME: Final[str] = "SKORE_HUB_URI"
URI: Final[str] = environ.get(ENV_VAR_NAME, DEFAULT)



"""Interface definition of the payload used to associate an artifact with a project."""

from abc import ABC, abstractmethod
from contextlib import AbstractContextManager, nullcontext
from functools import cached_property

from pydantic import BaseModel, ConfigDict, Field, computed_field

from skore_hub_project.artifact.upload import upload
from skore_hub_project.project.project import Project

Content = str | bytes | None


class Artifact(BaseModel, ABC):
    """
    Interface definition of the payload used to associate an artifact with a project.

    Attributes
    ----------
    project : Project
        The project to which the artifact's payload must be associated.
    content_type : str
        The content-type of the artifact content.

    Notes
    -----
    It triggers the upload of the content of the artifact, in a lazy way. It is uploaded
    as a file to the ``hub`` artifacts storage.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    project: Project = Field(repr=False, exclude=True)
    content_type: str = Field(init=False)

    @abstractmethod
    def content_to_upload(self) -> Content | AbstractContextManager[Content]:
        """
        Content of the artifact to upload.

        Example
        -------
        You can implement this ``abstractmethod`` to return directly the content:

            def content_to_upload(self) -> str:
                return "<str>"

        or to yield the content, as a ``contextmanager`` would:

            from contextlib import contextmanager

            @contextmanager
            def content_to_upload(self) -> Generator[str, None, None]:
                yield "<str>"
        """

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def checksum(self) -> str | None:
        """Checksum used to identify the content of the artifact."""
        contextmanager = self.content_to_upload()

        if not isinstance(contextmanager, AbstractContextManager):
            contextmanager = nullcontext(contextmanager)

        with contextmanager as content:
            if content is not None:
                return upload(
                    project=self.project,
                    content=content,
                    content_type=self.content_type,
                )

        return None



"""Class definition of the payload used to send an estimator report to ``hub``."""

from typing import ClassVar

from skore_hub_project.artifact.media import (
    Coefficients,
    EstimatorHtmlRepr,
    ImpurityDecrease,
    PermutationImportanceTest,
    PermutationImportanceTrain,
    PrecisionRecallTest,
    PrecisionRecallTrain,
    PredictionErrorTest,
    PredictionErrorTrain,
    RocTest,
    RocTrain,
    TableReportTest,
    TableReportTrain,
)
from skore_hub_project.artifact.media.media import Media
from skore_hub_project.metric import (
    AccuracyTest,
    AccuracyTrain,
    BrierScoreTest,
    BrierScoreTrain,
    FitTime,
    LogLossTest,
    LogLossTrain,
    PrecisionTest,
    PrecisionTrain,
    PredictTimeTest,
    PredictTimeTrain,
    R2Test,
    R2Train,
    RecallTest,
    RecallTrain,
    RmseTest,
    RmseTrain,
    RocAucTest,
    RocAucTrain,
)
from skore_hub_project.metric.metric import Metric
from skore_hub_project.protocol import EstimatorReport
from skore_hub_project.report.report import ReportPayload


class EstimatorReportPayload(ReportPayload[EstimatorReport]):
    """
    Payload used to send an estimator report to ``hub``.

    Attributes
    ----------
    METRICS : ClassVar[tuple[Metric, ...]]
        The metric classes that have to be computed from the report.
    MEDIAS : ClassVar[tuple[Media, ...]]
        The media classes that have to be computed from the report.
    project : Project
        The project to which the report payload should be sent.
    report : EstimatorReport
        The report on which to calculate the payload to be sent.
    key : str
        The key to associate to the report.
    """

    METRICS: ClassVar[tuple[type[Metric[EstimatorReport]], ...]] = (
        AccuracyTest,
        AccuracyTrain,
        BrierScoreTest,
        BrierScoreTrain,
        LogLossTest,
        LogLossTrain,
        PrecisionTest,
        PrecisionTrain,
        R2Test,
        R2Train,
        RecallTest,
        RecallTrain,
        RmseTest,
        RmseTrain,
        RocAucTest,
        RocAucTrain,
        # timings must be calculated last, or predictions must be cached before
        FitTime,
        PredictTimeTest,
        PredictTimeTrain,
    )
    MEDIAS: ClassVar[tuple[type[Media[EstimatorReport]], ...]] = (
        Coefficients,
        EstimatorHtmlRepr,
        ImpurityDecrease,
        PermutationImportanceTest,
        PermutationImportanceTrain,
        PrecisionRecallTest,
        PrecisionRecallTrain,
        PredictionErrorTest,
        PredictionErrorTrain,
        RocTest,
        RocTrain,
        TableReportTest,
        TableReportTrain,
    )



"""Definition of the payload used to associate a model category media with report."""

from typing import Literal

from skore_hub_project.artifact.media.media import Media, Report


class EstimatorHtmlRepr(Media[Report]):  # noqa: D101
    name: Literal["estimator_html_repr"] = "estimator_html_repr"
    data_source: None = None
    content_type: Literal["text/html"] = "text/html"

    def content_to_upload(self) -> str:  # noqa: D102
        import sklearn.utils

        estimator_html_repr: str = sklearn.utils.estimator_html_repr(
            self.report.estimator
        )

        return estimator_html_repr



"""Definition of the payload used to associate the pickled report with the report."""

from collections.abc import Generator
from contextlib import contextmanager
from io import BytesIO
from typing import Literal

from joblib import dump
from pydantic import Field

from skore_hub_project.artifact.artifact import Artifact
from skore_hub_project.protocol import CrossValidationReport, EstimatorReport

Report = EstimatorReport | CrossValidationReport


class Pickle(Artifact):
    """
    Payload used to associate the pickled report with the report, using ``joblib``.

    Attributes
    ----------
    project : Project
        The project to which the artifact's payload must be associated.
    content_type : str
        The content-type of the artifact content.
    report : EstimatorReport | CrossValidationReport
        The report to pickled.

    Notes
    -----
    It uploads the pickled report to the artifacts storage in a lazy way.

    The report is uploaded without its cache, to avoid salting the checksum.
    The report is primarily pickled on disk to reduce RAM footprint.
    """

    report: Report = Field(repr=False, exclude=True)
    content_type: Literal["application/octet-stream"] = "application/octet-stream"

    @contextmanager
    def content_to_upload(self) -> Generator[bytes, None, None]:
        """
        Content of the pickled report.

        Notes
        -----
        The report is pickled without its cache, to avoid salting the checksum.
        """
        reports = [self.report] + getattr(self.report, "estimator_reports_", [])
        caches = [report_to_clear._cache for report_to_clear in reports]

        self.report.clear_cache()

        try:
            with BytesIO() as stream:
                dump(self.report, stream)

                pickle_bytes = stream.getvalue()

            yield pickle_bytes
        finally:
            for report, cache in zip(reports, caches, strict=True):
                report._cache = cache



"""Class definition of the payload used to send a log loss metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric


class LogLoss(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.log_loss"
    name: str = "log_loss"
    verbose_name: str = "Log loss"
    greater_is_better: bool = False
    position: int = 4


class LogLossTrain(LogLoss):  # noqa: D101
    data_source: Literal["train"] = "train"


class LogLossTest(LogLoss):  # noqa: D101
    data_source: Literal["test"] = "test"


class LogLossMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.log_loss"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: str = "log_loss_mean"
    verbose_name: str = "Log loss - MEAN"
    greater_is_better: bool = False
    position: int = 4


class LogLossTrainMean(LogLossMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class LogLossTestMean(LogLossMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class LogLossStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.log_loss"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: str = "log_loss_std"
    verbose_name: str = "Log loss - STD"
    greater_is_better: bool = False
    position: None = None


class LogLossTrainStd(LogLossStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class LogLossTestStd(LogLossStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Definition of the payload used to associate a data category media with report."""

from typing import Literal

from skore_hub_project import switch_mpl_backend
from skore_hub_project.artifact.media.media import Media, Report
from skore_hub_project.protocol import EstimatorReport


class TableReport(Media[Report]):  # noqa: D101
    name: Literal["table_report"] = "table_report"
    data_source: Literal["train", "test"] | None = None
    content_type: Literal["application/vnd.skrub.table-report.v1+json"] = (
        "application/vnd.skrub.table-report.v1+json"
    )

    def content_to_upload(self) -> bytes:  # noqa: D102
        import orjson

        with switch_mpl_backend():
            display = (
                self.report.data.analyze()
                if self.data_source is None
                else self.report.data.analyze(data_source=self.data_source)
            )

        table_report = display.summary

        # Replace full dataset by its head/tail
        dataframe = table_report.pop("dataframe")
        table_report["extract_head"] = dataframe.head(3).to_dict(orient="split")
        table_report["extract_tail"] = dataframe.tail(3).to_dict(orient="split")

        # Remove irrelevant information
        del table_report["sample_table"]

        return orjson.dumps(
            table_report,
            option=(orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY),
        )


class TableReportTrain(TableReport[EstimatorReport]):  # noqa: D101
    data_source: Literal["train"] = "train"


class TableReportTest(TableReport[EstimatorReport]):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Token used for ``skore hub`` authentication."""

from __future__ import annotations

from datetime import datetime, timezone
from threading import RLock
from time import sleep
from urllib.parse import urljoin
from webbrowser import open as open_webbrowser

from httpx import HTTPStatusError, TimeoutException
from rich.align import Align
from rich.live import Live
from rich.panel import Panel

from skore_hub_project import console
from skore_hub_project.authentication.uri import URI


def get_oauth_device_login(success_uri: str | None = None) -> tuple[str, str, str]:
    """
    Initiate device OAuth flow.

    Initiates the OAuth device flow.
    Provides the user with a URL and a OTP code to authenticate the device.

    Parameters
    ----------
    success_uri : str, optional
        The URI to redirect to after successful authentication.
        If not provided, defaults to None.

    Returns
    -------
    tuple
        A tuple containing:
        - authorization_url: str
            The URL to which the user needs to navigate
        - device_code: str
            The device code used for authentication
        - user_code: str
            The user code that needs to be entered on the authorization page
    """
    from skore_hub_project.client.client import Client

    url = "identity/oauth/device/login"
    params = {"success_uri": success_uri} if success_uri is not None else {}

    with Client() as client:
        response = client.get(urljoin(URI, url), params=params).json()

        return (
            response["authorization_url"],
            response["device_code"],
            response["user_code"],
        )


def get_oauth_device_code_probe(device_code: str, *, timeout: int = 600) -> None:
    """
    Ensure authorization code is acknowledged.

    Start polling, wait for the authorization code to be acknowledged by the hub.
    This is mandatory to be authorize to exchange with a token.

    Parameters
    ----------
    device_code : str
        The device code to exchange for tokens.
    """
    from skore_hub_project.client.client import Client

    url = "identity/oauth/device/code-probe"
    params = {"device_code": device_code}

    with Client() as client:
        start = datetime.now()

        while True:
            try:
                client.get(urljoin(URI, url), params=params)
            except HTTPStatusError as exc:
                if exc.response.status_code != 400:
                    raise

                if (datetime.now() - start).total_seconds() >= timeout:
                    raise TimeoutException("Authentication timeout") from exc

                sleep(0.5)
            else:
                break


def post_oauth_device_callback(state: str, user_code: str) -> None:
    """
    Validate the user-provided device code.

    This endpoint verifies the code entered by the user during the device auth flow.

    Parameters
    ----------
    state: str
        The unique value identifying the device flow.
    user_code: str
        The code entered by the user.
    """
    from skore_hub_project.client.client import Client

    url = "identity/oauth/device/callback"
    data = {"state": state, "user_code": user_code}

    with Client() as client:
        client.post(urljoin(URI, url), data=data)


def get_oauth_device_token(device_code: str) -> tuple[str, str, str]:
    """
    Exchanges the device code for an access token.

    This endpoint completes the device authorization flow
    by exchanging the validated device code for an access token.

    Parameters
    ----------
    device_code : str
        The device code to exchange for tokens

    Returns
    -------
    tuple
        A tuple containing:
        - access_token : str
            The OAuth access token
        - refresh_token : str
            The OAuth refresh token
        - expires_at : str
            The expiration datetime as ISO 8601 str of the access token
    """
    from skore_hub_project.client.client import Client

    url = "identity/oauth/device/token"
    params = {"device_code": device_code}

    with Client() as client:
        response = client.get(urljoin(URI, url), params=params).json()
        tokens = response["token"]

        return (
            tokens["access_token"],
            tokens["refresh_token"],
            tokens["expires_at"],
        )


def post_oauth_refresh_token(refresh_token: str) -> tuple[str, str, str]:
    """
    Refresh an access token using a provided refresh token.

    This endpoint allows a client to obtain a new access token
    by providing a valid refresh token.

    Parameters
    ----------
    refresh_token : str
        A valid refresh token

    Returns
    -------
    tuple
        A tuple containing:
        - access_token : str
            The OAuth access token
        - refresh_token : str
            The OAuth refresh token
        - expires_at : str
            The expiration datetime as ISO 8601 str of the access token
    """
    from skore_hub_project.client.client import Client

    url = "identity/oauth/token/refresh"
    json = {"refresh_token": refresh_token}

    with Client() as client:
        response = client.post(urljoin(URI, url), json=json).json()

        return (
            response["access_token"],
            response["refresh_token"],
            response["expires_at"],
        )


class Token:
    """
    Token used for ``skore hub`` authentication, as HTTP header parameters.

    Notes
    -----
    Refresh the token on-the-fly if necessary.
    """

    def __init__(self, *, timeout: int = 600, live: Live | None = None) -> None:
        url, device_code, user_code = get_oauth_device_login()
        panel = Panel(
            Align.center(
                "[b]API key not detected.[/b]\n\n"
                "Starting interactive authentication for the session.\n"
                "[i]We recommend that you create an API key and use it to log in, "
                "at [link=https://skore.probabl.ai/account]"
                "https://skore.probabl.ai/account[/link].[/i]\n\n"
                "Opening browser for interactive authentication; if this fails, "
                f"please visit:\n[link={url}]{url}[/link]"
            ),
            title="[cyan]Login to [bold]Skore Hub",
            border_style="cyan",
            padding=1,
        )

        if live:
            live.update(panel)
            live.refresh()
        else:
            console.print(panel, soft_wrap=True)

        open_webbrowser(url)

        get_oauth_device_code_probe(device_code, timeout=timeout)
        post_oauth_device_callback(device_code, user_code)

        access, refreshment, expiration = get_oauth_device_token(device_code)

        self.__lock = RLock()
        self.__access = access
        self.__refreshment = refreshment
        self.__expiration = datetime.fromisoformat(expiration)

    def __call__(self) -> dict[str, str]:  # noqa: D102
        with self.__lock:
            if self.__expiration <= datetime.now(timezone.utc):
                access, refreshment, expiration = post_oauth_refresh_token(
                    self.__refreshment
                )

                self.__access = access
                self.__refreshment = refreshment
                self.__expiration = datetime.fromisoformat(expiration)

            return {"Authorization": f"Bearer {self.__access}"}



"""Definition of the payload used to associate a performance media with report."""

from __future__ import annotations

from abc import ABC
from collections.abc import Callable
from functools import reduce
from io import BytesIO
from typing import ClassVar, Literal, cast

from matplotlib import pyplot as plt

from skore_hub_project import switch_mpl_backend
from skore_hub_project.artifact.media.media import Media, Report
from skore_hub_project.protocol import Display


class Performance(Media[Report], ABC):  # noqa: D101
    accessor: ClassVar[str]
    content_type: Literal["image/svg+xml"] = "image/svg+xml"

    def content_to_upload(self) -> bytes | None:  # noqa: D102
        try:
            function = cast(
                "Callable[..., Display]",
                reduce(getattr, self.accessor.split("."), self.report),
            )
        except AttributeError:
            return None

        display = (
            function()
            if self.data_source is None
            else function(data_source=self.data_source)
        )

        with switch_mpl_backend(), BytesIO() as stream:
            display.plot()
            display.figure_.savefig(stream, format="svg", bbox_inches="tight")  # type: ignore[attr-defined]
            plt.close(display.figure_)  # type: ignore[attr-defined]

            figure_bytes = stream.getvalue()

        return figure_bytes


class PrecisionRecall(Performance[Report], ABC):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision_recall"
    name: Literal["precision_recall"] = "precision_recall"


class PrecisionRecallTrain(PrecisionRecall[Report]):  # noqa: D101
    data_source: Literal["train"] = "train"


class PrecisionRecallTest(PrecisionRecall[Report]):  # noqa: D101
    data_source: Literal["test"] = "test"


class PredictionError(Performance[Report], ABC):  # noqa: D101
    accessor: ClassVar[str] = "metrics.prediction_error"
    name: Literal["prediction_error"] = "prediction_error"


class PredictionErrorTrain(PredictionError[Report]):  # noqa: D101
    data_source: Literal["train"] = "train"


class PredictionErrorTest(PredictionError[Report]):  # noqa: D101
    data_source: Literal["test"] = "test"


class Roc(Performance[Report], ABC):  # noqa: D101
    accessor: ClassVar[str] = "metrics.roc"
    name: Literal["roc"] = "roc"


class RocTrain(Roc[Report]):  # noqa: D101
    data_source: Literal["train"] = "train"


class RocTest(Roc[Report]):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Class definition of the payload used to send a brier score metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric


class BrierScore(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.brier_score"]] = "metrics.brier_score"
    name: Literal["brier_score"] = "brier_score"
    verbose_name: Literal["Brier score"] = "Brier score"
    greater_is_better: Literal[False] = False
    position: None = None


class BrierScoreTrain(BrierScore):  # noqa: D101
    data_source: Literal["train"] = "train"


class BrierScoreTest(BrierScore):  # noqa: D101
    data_source: Literal["test"] = "test"


class BrierScoreMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.brier_score"]] = "metrics.brier_score"
    aggregate: ClassVar[Literal["mean"]] = "mean"
    name: Literal["brier_score_mean"] = "brier_score_mean"
    verbose_name: Literal["Brier score - MEAN"] = "Brier score - MEAN"
    greater_is_better: Literal[False] = False
    position: None = None


class BrierScoreTrainMean(BrierScoreMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class BrierScoreTestMean(BrierScoreMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class BrierScoreStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[Literal["metrics.brier_score"]] = "metrics.brier_score"
    aggregate: ClassVar[Literal["std"]] = "std"
    name: Literal["brier_score_std"] = "brier_score_std"
    verbose_name: Literal["Brier score - STD"] = "Brier score - STD"
    greater_is_better: Literal[False] = False
    position: None = None


class BrierScoreTrainStd(BrierScoreStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class BrierScoreTestStd(BrierScoreStd):  # noqa: D101
    data_source: Literal["test"] = "test"



"""Protocols definition used to remove adherence to ``skore`` for type checking."""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Literal, Protocol, runtime_checkable

if TYPE_CHECKING:
    from pandas import DataFrame

DataSource = Literal["train", "test"] | None
Aggregate = Literal["mean", "std"]


@runtime_checkable
class Display(Protocol):
    """Protocol equivalent to ``skore.Display``."""

    frame: Callable[..., DataFrame]
    plot: Callable[..., None]


@runtime_checkable
class EstimatorReport(Protocol):
    """Protocol equivalent to ``skore.EstimatorReport``."""

    _hash: int
    cache_predictions: Any
    clear_cache: Any
    _cache: Any
    metrics: Any
    data: Any
    ml_task: str
    estimator: Any
    estimator_name_: str
    X_train: Any
    y_train: Any
    X_test: Any
    y_test: Any
    fit: Any


@runtime_checkable
class CrossValidationReport(Protocol):
    """Protocol equivalent to ``skore.CrossValidationReport``."""

    _hash: int
    cache_predictions: Any
    clear_cache: Any
    _cache: Any
    metrics: Any
    data: Any
    estimator_reports_: Any
    ml_task: str
    estimator: Any
    estimator_name_: str
    X: Any
    y: Any
    splitter: Any
    split_indices: Any



"""Class definition of the payload used to send a cross-validation report to ``hub``."""

from collections import Counter, defaultdict
from functools import cached_property
from typing import Any, ClassVar

import numpy as np
from pydantic import computed_field
from scipy.stats import gaussian_kde
from sklearn.model_selection._split import _CVIterableWrapper

from skore_hub_project.artifact.media import (
    EstimatorHtmlRepr,
    PrecisionRecallTest,
    PrecisionRecallTrain,
    PredictionErrorTest,
    PredictionErrorTrain,
    RocTest,
    RocTrain,
)
from skore_hub_project.artifact.media.data import TableReport
from skore_hub_project.artifact.media.media import Media
from skore_hub_project.metric import (
    AccuracyTestMean,
    AccuracyTestStd,
    AccuracyTrainMean,
    AccuracyTrainStd,
    BrierScoreTestMean,
    BrierScoreTestStd,
    BrierScoreTrainMean,
    BrierScoreTrainStd,
    FitTimeMean,
    FitTimeStd,
    LogLossTestMean,
    LogLossTestStd,
    LogLossTrainMean,
    LogLossTrainStd,
    PrecisionTestMean,
    PrecisionTestStd,
    PrecisionTrainMean,
    PrecisionTrainStd,
    PredictTimeTestMean,
    PredictTimeTestStd,
    PredictTimeTrainMean,
    PredictTimeTrainStd,
    R2TestMean,
    R2TestStd,
    R2TrainMean,
    R2TrainStd,
    RecallTestMean,
    RecallTestStd,
    RecallTrainMean,
    RecallTrainStd,
    RmseTestMean,
    RmseTestStd,
    RmseTrainMean,
    RmseTrainStd,
    RocAucTestMean,
    RocAucTestStd,
    RocAucTrainMean,
    RocAucTrainStd,
)
from skore_hub_project.metric.metric import Metric
from skore_hub_project.protocol import CrossValidationReport
from skore_hub_project.report.estimator_report import EstimatorReportPayload
from skore_hub_project.report.report import ReportPayload


class CrossValidationReportPayload(ReportPayload[CrossValidationReport]):
    """
    Payload used to send a cross-validation report to ``hub``.

    Attributes
    ----------
    METRICS : ClassVar[tuple[Metric, ...]]
        The metric classes that have to be computed from the report.
    MEDIAS : ClassVar[tuple[Media, ...]]
        The media classes that have to be computed from the report.
    project : Project
        The project to which the report payload should be sent.
    report : CrossValidationReport
        The report on which to calculate the payload to be sent.
    key : str
        The key to associate to the report.
    """

    METRICS: ClassVar[tuple[type[Metric[CrossValidationReport]], ...]] = (
        AccuracyTestMean,
        AccuracyTestStd,
        AccuracyTrainMean,
        AccuracyTrainStd,
        BrierScoreTestMean,
        BrierScoreTestStd,
        BrierScoreTrainMean,
        BrierScoreTrainStd,
        LogLossTestMean,
        LogLossTestStd,
        LogLossTrainMean,
        LogLossTrainStd,
        PrecisionTestMean,
        PrecisionTestStd,
        PrecisionTrainMean,
        PrecisionTrainStd,
        R2TestMean,
        R2TestStd,
        R2TrainMean,
        R2TrainStd,
        RecallTestMean,
        RecallTestStd,
        RecallTrainMean,
        RecallTrainStd,
        RmseTestMean,
        RmseTestStd,
        RmseTrainMean,
        RmseTrainStd,
        RocAucTestMean,
        RocAucTestStd,
        RocAucTrainMean,
        RocAucTrainStd,
        # timings must be calculated last, or predictions must be cached before
        FitTimeMean,
        FitTimeStd,
        PredictTimeTestMean,
        PredictTimeTestStd,
        PredictTimeTrainMean,
        PredictTimeTrainStd,
    )
    MEDIAS: ClassVar[tuple[type[Media[CrossValidationReport]], ...]] = (
        EstimatorHtmlRepr,
        PrecisionRecallTest,
        PrecisionRecallTrain,
        PredictionErrorTest,
        PredictionErrorTrain,
        RocTest,
        RocTrain,
        TableReport,
    )

    def model_post_init(self, _: Any) -> None:  # noqa: D102
        self.__sample_to_class_index: list[int] | None
        self.__classes: list[str] | None

        if "classification" in self.ml_task:
            class_to_class_indice: defaultdict[Any, int] = defaultdict(
                lambda: len(class_to_class_indice)
            )

            self.__sample_to_class_index = [
                class_to_class_indice[sample] for sample in self.report.y
            ]

            assert len(self.__sample_to_class_index) == len(self.report.X)

            self.__classes = [str(class_) for class_ in class_to_class_indice]

            assert max(self.__sample_to_class_index) == (len(self.__classes) - 1)
        else:
            self.__sample_to_class_index = None
            self.__classes = None

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def dataset_size(self) -> int:
        """Size of the dataset."""
        return len(self.report.X)

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def splitting_strategy(self) -> dict[str, Any]:
        """
        Splitting strategy used to split the dataset into train and test sets.

        This includes the number of splits, the number of repeats, the seed,
        and the distribution of the train and test sets.

        The distribution of each split is computed by dividing the split into a maximum
        of 200 buckets, and averaging the number of samples belonging to the test-set in
        each of these buckets. @TODO: find a better representation of the distribution.
        """
        splits = []

        for train_indices, test_indices in self.report.split_indices:
            train_y = self.report.y[train_indices]
            test_y = self.report.y[test_indices]
            train_target_distribution: list[float] = []
            test_target_distribution: list[float] = []

            if self.__classes:
                train = {str(label): count for label, count in Counter(train_y).items()}
                test = {str(label): count for label, count in Counter(test_y).items()}

                for label in self.__classes:
                    train_target_distribution.append(train.get(label, 0))
                    test_target_distribution.append(test.get(label, 0))

            else:
                linspace = np.linspace(
                    float(train_y.min()), float(train_y.max()), num=100
                )
                train_kernel = gaussian_kde(train_y)
                train_target_distribution = [float(x) for x in train_kernel(linspace)]
                test_kernel = gaussian_kde(test_y)
                test_target_distribution = [float(x) for x in test_kernel(linspace)]

            # remove this when a better solution is found
            # see #2212 for details
            buckets_number = min(len(self.report.X), 200)
            split = np.zeros(len(self.report.X), dtype=int)
            split[test_indices] = 1

            splits.append(
                {
                    "train": {
                        "sample_count": len(train_indices),
                        "target_distribution": train_target_distribution,
                        "groups": None,
                    },
                    "test": {
                        "sample_count": len(test_indices),
                        "target_distribution": test_target_distribution,
                        "groups": None,
                    },
                    "train_test_distribution": [
                        float(np.mean(bucket))
                        for bucket in np.array_split(split, buckets_number)
                    ],
                }
            )

        return {
            "strategy_name": (
                isinstance(self.report.splitter, _CVIterableWrapper)
                and "custom"
                or self.report.splitter.__class__.__name__
            ),
            "repeat_count": getattr(self.report.splitter, "n_repeats", 1),
            "seed": str(getattr(self.report.splitter, "random_state", "")),
            "splits": splits,
        }

    groups: list[int] | None = None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def class_names(self) -> list[str] | None:
        """In classification, the class names of the dataset used in the report."""
        return self.__classes

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def target_range(self) -> list[float] | None:
        """The range of the target values of the dataset used in the report."""
        if self.__classes:
            return None
        return [float(self.report.y.min()), float(self.report.y.max())]

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def estimators(self) -> list[EstimatorReportPayload]:
        """The estimators used in each split by the report in a payload format."""
        return [
            EstimatorReportPayload(
                project=self.project,
                report=report,
                key=f"{self.key}:estimator-report",
            )
            for report in self.report.estimator_reports_
        ]



"""Function definition of the content ``Serializer``."""

from __future__ import annotations

from functools import cached_property
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any

from blake3 import blake3 as Blake3


class Serializer:
    """Serialize a content directly on disk to reduce RAM footprint."""

    def __init__(self, content: str | bytes):
        if isinstance(content, str):
            self.filepath.write_text(content, encoding="utf-8")
        else:
            self.filepath.write_bytes(content)

    def __enter__(self) -> Serializer:  # noqa: D105
        return self

    def __exit__(self, *args: Any) -> None:  # noqa: D105
        self.filepath.unlink(True)

    @cached_property
    def filepath(self) -> Path:
        """The filepath used to serialize the content."""
        with NamedTemporaryFile(mode="w+b", delete=False) as file:
            return Path(file.name)

    @cached_property
    def checksum(self) -> str:
        """
        The checksum of the serialized content.

        Notes
        -----
        Depending on the size of the serialized content, the checksum can be computed on
        one or more threads:

            Note that this can be slower for inputs shorter than ~1 MB

        https://github.com/oconnor663/blake3-py
        """
        from skore_hub_project import bytes_to_b64_str

        hasher = Blake3(max_threads=(1 if self.size < 1e6 else Blake3.AUTO))
        checksum = hasher.update_mmap(self.filepath).digest()

        return f"blake3-{bytes_to_b64_str(checksum)}"

    @cached_property
    def size(self) -> int:
        """The size of the serialized content, in bytes."""
        return self.filepath.stat().st_size



"""Hatchling hooks used to dynamically update the metadata of the package."""

from pathlib import Path
from shutil import copy

from hatchling.metadata.plugin.interface import MetadataHookInterface


def readlines(filepath):
    """Yied the content of ``filepath`` line by line."""
    with open(filepath) as file:
        yield from file


class MetadataHook(MetadataHookInterface):
    """Hatchling hooks used to dynamically update the metadata of the package."""

    def update(self, metadata):
        """
        Update the metadata of the package, after it has been loaded.

        Update ``LICENSE`` and ``README`` from root files, shared in the GH repository.
        Update ``VERSION`` from environment variables filled by the GH release pipeline.

        Notes
        -----
        Example of configuration:

            {
                'path': 'hatch/metadata.py',
                'version-default': '0.0.0+unknown',
                'license': {'file': '../LICENSE'},
                'readme': {'file': '../README.md'},
            }
        """
        # Copy LICENCE file from root to `sdist`
        license_filepath = Path(self.root, self.config["license"]["file"])
        license_filename = license_filepath.name

        copy(license_filepath, self.root)

        # Retrieve README from root files
        readme_filepath = self.config["readme"]["file"]
        readme = Path(self.root, readme_filepath).read_text(encoding="utf-8")

        # Retrieve VERSION from file created by the CI
        try:
            version = Path(self.root, "VERSION.txt").read_text(encoding="utf-8")
        except FileNotFoundError:
            version = self.config["version-default"]

        # Update metadata
        metadata["license-files"] = [license_filename]
        metadata["readme"] = {"text": readme, "content-type": "text/markdown"}
        metadata["version"] = version



"""Class definition of the payload used to send a precision metric to ``hub``."""

from __future__ import annotations

from typing import ClassVar, Literal

from .metric import CrossValidationReportMetric, EstimatorReportMetric, cast_to_float


class Precision(EstimatorReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision"
    name: str = "precision"
    verbose_name: str = "Precision (macro)"
    greater_is_better: bool = True
    position: None = None

    def compute(self) -> None:
        """Compute the value of the metric."""
        try:
            function = self.report.metrics.precision
        except AttributeError:
            self.value = None
        else:
            self.value = cast_to_float(
                function(data_source=self.data_source, average="macro")
            )


class PrecisionTrain(Precision):  # noqa: D101
    data_source: Literal["train"] = "train"


class PrecisionTest(Precision):  # noqa: D101
    data_source: Literal["test"] = "test"


class PrecisionMean(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision"
    name: str = "precision_mean"
    verbose_name: str = "Precision (macro) - MEAN"
    greater_is_better: bool = True
    position: None = None

    def compute(self) -> None:
        """Compute the value of the metric."""
        try:
            function = self.report.metrics.precision
        except AttributeError:
            self.value = None
        else:
            dataframe = function(
                data_source=self.data_source, aggregate="mean", average="macro"
            )
            self.value = cast_to_float(dataframe.iloc[0, 0])


class PrecisionTrainMean(PrecisionMean):  # noqa: D101
    data_source: Literal["train"] = "train"


class PrecisionTestMean(PrecisionMean):  # noqa: D101
    data_source: Literal["test"] = "test"


class PrecisionStd(CrossValidationReportMetric):  # noqa: D101
    accessor: ClassVar[str] = "metrics.precision"
    name: str = "precision_std"
    verbose_name: str = "Precision (macro) - STD"
    greater_is_better: bool = False
    position: None = None

    def compute(self) -> None:
        """Compute the value of the metric."""
        try:
            function = self.report.metrics.precision
        except AttributeError:
            self.value = None
        else:
            dataframe = function(
                data_source=self.data_source, aggregate="std", average="macro"
            )
            self.value = cast_to_float(dataframe.iloc[0, 0])


class PrecisionTrainStd(PrecisionStd):  # noqa: D101
    data_source: Literal["train"] = "train"


class PrecisionTestStd(PrecisionStd):  # noqa: D101
    data_source: Literal["test"] = "test"
